﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Linq.Expressions;
using System.Reflection;

namespace Achiral
{
    using Achiral.Extension;
    using System.Diagnostics;
    using System.Collections;
    using LinqExpression = System.Linq.Expressions.Expression;

    /// <see>http://www.ecma-international.org/publications/techreports/E-TR-089.htm</see>
    [Serializable]
    public enum Unit { Unit };

    public static partial class Make
    {
        /// <see>http://blogs.wankuma.com/mayuki/archive/2007/12/06/111901.aspx</see>
        public static Dictionary<string, T> Dictionary<T>(params Expression<Func<object, T>>[] exprs)
        {
            AssertNonNullArgument(exprs, "exprs");
            return exprs.ToDictionary(expr => expr.Parameters[0].Name, expr => expr.Compile().Invoke(null));
        }

        public static Action Action(Action action) { return action; }
        public static Action<T> Action<T>(Action<T> action) { return action; }
        public static Action<T1, T2> Action<T1, T2>(Action<T1, T2> action) { return action; }
        public static Action<T1, T2, T3> Action<T1, T2, T3>(Action<T1, T2, T3> action) { return action; }
        public static Action<T1, T2, T3, T4> Action<T1, T2, T3, T4>(Action<T1, T2, T3, T4> action) { return action; }

        public static Func<TResult> Func<TResult>(Func<TResult> func) { return func; }
        public static Func<T, TResult> Func<T, TResult>(Func<T, TResult> func) { return func; }
        public static Func<T1, T2, TResult> Func<T1, T2, TResult>(Func<T1, T2, TResult> func) { return func; }
        public static Func<T1, T2, T3, TResult> Func<T1, T2, T3, TResult>(Func<T1, T2, T3, TResult> func) { return func; }
        public static Func<T1, T2, T3, T4, TResult> Func<T1, T2, T3, T4, TResult>(Func<T1, T2, T3, T4, TResult> func) { return func; }

        public static Expression<Action> Expression(Expression<Action> expr) { return expr; }
        public static Expression<Action<T>> Expression<T>(Expression<Action<T>> expr) { return expr; }
        public static Expression<Action<T1, T2>> Expression<T1, T2>(Expression<Action<T1, T2>> expr) { return expr; }
        public static Expression<Action<T1, T2, T3>> Expression<T1, T2, T3>(Expression<Action<T1, T2, T3>> expr) { return expr; }
        public static Expression<Action<T1, T2, T3, T4>> Expression<T1, T2, T3, T4>(Expression<Action<T1, T2, T3, T4>> expr) { return expr; }
        
        public static Expression<Func<TResult>> Expression<TResult>(Expression<Func<TResult>> expr) { return expr; }
        public static Expression<Func<T1, TResult>> Expression<T1, TResult>(Expression<Func<T1, TResult>> expr) { return expr; }
        public static Expression<Func<T1, T2, TResult>> Expression<T1, T2, TResult>(Expression<Func<T1, T2, TResult>> expr) { return expr; }
        public static Expression<Func<T1, T2, T3, TResult>> Expression<T1, T2, T3, TResult>(Expression<Func<T1, T2, T3, TResult>> expr) { return expr; }
        public static Expression<Func<T1, T2, T3, T4, TResult>> Expression<T1, T2, T3, T4, TResult>(Expression<Func<T1, T2, T3, T4, TResult>> expr) { return expr; }

        public static Expression EqualityExpression(Expression left, Expression right)
        {
            var eq = LinqExpression.Equal(left, right);
            if (eq.Method != null)
            {
                return LinqExpression.Equal(left, right, eq.IsLiftedToNull, eq.Method);
            }

            var methodInfo =
                 (from method in eq.Left.Type.GetAllFlattenMethods()
                  where !method.IsStatic
                  where method.Name == "Equals"
                  where method.ReturnType == typeof(bool)
                  where method.GetParameters().Length == 1
                  let param = method.GetParameters()[0]
                  where param.ParameterType.IsAssignableFrom(eq.Left.Type)
                  let paramTypeLevel = eq.Left.Type.GetHierarchyLevel() - param.ParameterType.GetHierarchyLevel()
                  orderby paramTypeLevel ascending
                  select method).First();

            return LinqExpression.Call(left, methodInfo, right);
        }

        public static MethodInfo MethodInfo(Expression<Action> expr) { return expr.GetMethodInfo(); }
        public static MethodInfo MethodInfo<TResult>(Expression<Func<TResult>> expr) { return expr.GetMethodInfo(); }
        public static PropertyInfo PropertyInfo<TResult>(Expression<Func<TResult>> expr) { return expr.GetPropertyInfo(); }
        public static FieldInfo FieldInfo<TResult>(Expression<Func<TResult>> expr) { return expr.GetFieldInfo(); }
        public static ConstructorInfo ConstructorInfo<TResult>(Expression<Func<TResult>> expr) { return expr.GetConstructorInfo(); }

        public static T[] Array<T>(params T[] array) { return array; }

        public static IEnumerable<T> Sequence<T>(T element) { return Enumerable.Repeat(element, 1); }

        public static IEnumerable<T> Repeat<T>(T element) { while (true) yield return element; }
        public static IEnumerable<T> Repeat<T>(Func<T> elementInitializer) { return RepeatImpl<T>.Default.Repeat(elementInitializer); }
        public static IEnumerable<T> Cycle<T>(T item1, T item2)
        {
            while (true)
            {
                yield return item1;
                yield return item2;
            }
        }
        public static IEnumerable<T> Cycle<T>(T item1, T item2, T item3)
        {
            while (true)
            {
                yield return item1;
                yield return item2;
                yield return item3;
            }
        }
        public static IEnumerable<T> Cycle<T>(T item1, T item2, T item3, T item4)
        {
            while (true)
            {
                yield return item1;
                yield return item2;
                yield return item3;
                yield return item4;
            }
        }
        public static IEnumerable<T> Cycle<T>(params T[] source)
        {
            AssertNonNullArgument(source, "source");
            while (true)
                foreach (var item in source)
                {
                    yield return item;
                }
        }

        public static bool Equality<T>(T arg1, T arg2) { return EqualityComparer<T>.Default.Equals(arg1, arg2); }

        /// <see>http://blogs.msdn.com/madst/archive/2007/05/11/recursive-lambda-expressions.aspx</see>
        public static Func<T, T> Fix<T>(Func<Func<T, T>, Func<T, T>> func) { return Fun<T>.Fix(func); }
        /// <see>http://blogs.msdn.com/madst/archive/2007/05/11/recursive-lambda-expressions.aspx</see>
        public static Func<T, T> Fix<T>(Func<Func<T, T>, Func<T, T>> func, T dummy) { return Fun<T>.Fix(func); }
    }

    public static class Exactly<TAssert>
    {
        public static TAssert Is(TAssert arg)
        {
            return arg;
        }
        [Obsolete("Type Assertion Failure", true)]
        public static TArg Is<TArg>(TArg arg)
        {
            return arg;
        }
    }

    public static partial class Make
    {
        public static void Swap<T>(ref T left, ref T right)
        {
            T tmp = left;
            left = right;
            right = tmp;
        }
    }

    public static partial class Make
    {
        public static PermutationSource Permutation(int numberOfElements)
        {
            return new PermutationSource(numberOfElements);
        }
    }

    public static partial class Make
    {
        /// <see>http://blogs.msdn.com/madst/archive/2007/05/11/recursive-lambda-expressions.aspx</see>
        private static class Fun<T>
        {
            private delegate U SelfApplicable<U>(SelfApplicable<U> self);
            private static SelfApplicable<Func<Func<Func<T, T>, Func<T, T>>, Func<T, T>>> Y = y => f => x => f(y(y)(f))(x);
            public static Func<Func<Func<T, T>, Func<T, T>>, Func<T, T>> Fix = Y(Y);
        }

        private static MethodInfo GetMethodInfo(this Expression expr)
        {
            AssertNonNullArgument(expr, "expr");

            switch (expr.NodeType)
            {
                case ExpressionType.Lambda:
                    return ((expr as LambdaExpression).Body).GetMethodInfo();
                case ExpressionType.Quote:
                    return ((expr as UnaryExpression).Operand).GetMethodInfo();
                case ExpressionType.Call:
                    var childExpr = expr as MethodCallExpression;
                    return childExpr != null ? childExpr.Method : null;
                case ExpressionType.MemberAccess:
                    var me = expr as MemberExpression;
                    var pi = me.Member as PropertyInfo;
                    return pi != null ? pi.GetGetMethod(true) : null;
                default:
                    return null;
            }
        }

        private static ConstructorInfo GetConstructorInfo(this Expression expr)
        {
            AssertNonNullArgument(expr, "expr");

            switch (expr.NodeType)
            {
                case ExpressionType.Lambda:
                    return ((expr as LambdaExpression).Body).GetConstructorInfo();
                case ExpressionType.Quote:
                    return ((expr as UnaryExpression).Operand).GetConstructorInfo();
                case ExpressionType.New:
                    return (expr as NewExpression).Constructor;
                default:
                    return null;
            }
        }

        private static PropertyInfo GetPropertyInfo(this Expression expr)
        {
            AssertNonNullArgument(expr, "expr");

            switch (expr.NodeType)
            {
                case ExpressionType.Lambda:
                    return ((expr as LambdaExpression).Body).GetPropertyInfo();
                case ExpressionType.Quote:
                    return ((expr as UnaryExpression).Operand).GetPropertyInfo();
                case ExpressionType.MemberAccess:
                    var me = expr as MemberExpression;
                    var info = me.Member as PropertyInfo;
                    return info;
                default:
                    return null;
            }
        }

        private static FieldInfo GetFieldInfo(this Expression expr)
        {
            AssertNonNullArgument(expr, "expr");

            switch (expr.NodeType)
            {
                case ExpressionType.Lambda:
                    return ((expr as LambdaExpression).Body).GetFieldInfo();
                case ExpressionType.Quote:
                    return ((expr as UnaryExpression).Operand).GetFieldInfo();
                case ExpressionType.MemberAccess:
                    var me = expr as MemberExpression;
                    var info = me.Member as FieldInfo;
                    return info;
                default:
                    return null;
            }
        }
    }

    public static partial class Make
    {
        private class RepeatImpl<T>
        {
            public virtual IEnumerable<T> Repeat(Func<T> elementInitializer)
            {
                var element = elementInitializer();
                while (true)
                {
                    yield return element;
                }
            }
            public static readonly RepeatImpl<T> Default;
            static RepeatImpl()
            {
                var t = typeof(T);
                if (typeof(IDisposable).IsAssignableFrom(t))
                {
                    var implType = typeof(RepeatImplWithUsing<>).MakeGenericType(t);
                    Default = Activator.CreateInstance(implType) as RepeatImpl<T>;
                }
                else
                {
                    Default = new RepeatImpl<T>();
                }
            }
        }
        private sealed class RepeatImplWithUsing<U> : RepeatImpl<U> where U : IDisposable
        {
            public override IEnumerable<U> Repeat(Func<U> elementInitializer)
            {
                using (var element = elementInitializer())
                {
                    while (true)
                    {
                        yield return element;
                    }
                }
            }
        }
    }

    public static partial class Make
    {
        [DebuggerStepThroughAttribute]
        internal static void AssertNonNullArgument<T>(T argument, string argumentName)
        {
            if (argument == null) throw new ArgumentNullException(argumentName);
        }
    }

    namespace Extension
    {
        public static partial class MakeWorld
        {
            [DebuggerStepThroughAttribute]
            internal static void AssertNonNullArgument<T>(T argument, string argumentName)
            {
                if(argument == null) throw new ArgumentNullException(argumentName);
            }
        }
    }
}

