﻿using System;
using System.Linq;
using System.Diagnostics;

namespace Achiral
{
    using Achiral.Extension;
    using System.ComponentModel;
    using System.Runtime.InteropServices;

    public static partial class Make
    {
        public static Maybe<T> Maybe<T>(T value)
        {
            return new Maybe<T>(value);
        }
        public static Maybe<T> Maybe<T>(T? value) where T : struct
        {
            return value.HasValue ? new Maybe<T>(value.Value) : new Maybe<T>();
        }
        public static T? ToNullable<T>(this Maybe<T> maybeValue) where T : struct
        {
            return maybeValue.IsNothing ? new T?() : maybeValue.Unwrap();
        }
    }

    [DebuggerDisplay("{IsNothing ? (object)\"Nothing\" : (object)Unwrap()}")]
    [StructLayout(LayoutKind.Auto)]
    public partial struct Maybe<T>
    {
        private readonly T Value;
        private bool _hasValue;
        public Maybe(T value)
        {
            Value = value;
            _hasValue = (value != null);
        }
        public bool IsNothing { get { return !_hasValue; } }
        public T Unwrap() { return Value; }
        public static explicit operator T(Maybe<T> maybe) { return maybe.Unwrap(); }

        public Maybe<U> Select<U>(Func<T, U?> func) where U : struct
        {
            return this.IsNothing ? new Maybe<U>() : Make.Maybe<U>(func(this.Unwrap()));
        }
        public Maybe<U> Select<U>(Func<T, U> func)
        {
            return this.IsNothing ? new Maybe<U>() : Make.Maybe<U>(func(this.Unwrap()));
        }
        public Maybe<U> SelectMany<U>(Func<T, Maybe<U>> func)
        {
            return this.IsNothing ? new Maybe<U>() : func(this.Unwrap());
        }
        public Maybe<V> SelectMany<U, V>(Func<T, Maybe<U>> func, Func<T, U, V> resultSelector)
        {
            if (this.IsNothing) return new Maybe<V>();
            var rawT = this.Unwrap();
            var temp = func(rawT);
            return temp.IsNothing ? new Maybe<V>() : new Maybe<V>(resultSelector(rawT, temp.Unwrap()));
        }
        public override string ToString()
        {
            return this.IsNothing ? "[" + typeof(T).Name + "] Nothing" : ((T)this).ToString();
        }
        public T GetValueOrDefault()
        {
            return this.IsNothing ? default(T) : this.Unwrap();
        }
        public T GetValueOrDefault(T defaultValue)
        {
            return this.IsNothing ? defaultValue : this.Unwrap();
        }
    }

    namespace Extension
    {
        using System.Collections.Generic;
        public static partial class MakeWorld
        {
            public static IEnumerable<T> Normalize<T>(this Maybe<IEnumerable<T>> maybeSource)
            {
                return maybeSource.IsNothing ? Enumerable.Empty<T>() : maybeSource.Unwrap();
            }
            public static T[] Normalize<T>(this Maybe<T[]> maybeSource)
            {
                return maybeSource.IsNothing ? Make.Array<T>() : maybeSource.Unwrap();
            }
        }
    }
}