﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Collections;
using System.Runtime.InteropServices;

namespace Achiral.Extension
{
    public static partial class MakeWorld
    {
        public static IEnumerable<TResult> Pairwise<T, TResult>(this IEnumerable<T> source, Func<T, T, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(selector, "selector");

            using (var enumerator = source.GetEnumerator())
            {
                if (!enumerator.MoveNext())
                {
                    yield break;
                }
                var prev = enumerator.Current;
                while (enumerator.MoveNext())
                {
                    yield return selector(prev, enumerator.Current);
                    prev = enumerator.Current;
                }
            }
        }

        public static IEnumerable<TResult> Select<T, TResult>(this T[,] source, Func<T, int, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(selector, "selector");

            checked
            {
                for (int i0 = 0; i0 < source.GetLength(0); ++i0)
                {
                    for (int i1 = 0; i1 < source.GetLength(1); ++i1)
                    {
                        yield return selector(source[i0, i1], i0, i1);
                    }
                }
            }
        }
#if !XBOX360
        public static IEnumerable<TResult> LongSelect<T, TResult>(this T[,] source, Func<T, long, long, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(selector, "selector");

            checked
            {
                for (long i0 = 0; i0 < source.GetLongLength(0); ++i0)
                {
                    for (long i1 = 0; i1 < source.GetLongLength(1); ++i1)
                    {
                        yield return selector(source[i0, i1], i0, i1);
                    }
                }
            }
        }
#endif
        public static IEnumerable<TResult> Select<T, TResult>(this T[, ,] source, Func<T, int, int, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(selector, "selector");

            checked
            {
                for (int i0 = 0; i0 < source.GetLength(0); ++i0)
                {
                    for (int i1 = 0; i1 < source.GetLength(1); ++i1)
                    {
                        for (int i2 = 0; i2 < source.GetLength(2); ++i2)
                        {
                            yield return selector(source[i0, i1, i2], i0, i1, i2);
                        }
                    }
                }
            }
        }
#if !XBOX360
        public static IEnumerable<TResult> LongSelect<T, TResult>(this T[, ,] source, Func<T, long, long, long, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(selector, "selector");

            checked
            {
                for (long i0 = 0; i0 < source.GetLongLength(0); ++i0)
                {
                    for (long i1 = 0; i1 < source.GetLongLength(1); ++i1)
                    {
                        for (long i2 = 0; i2 < source.GetLongLength(2); ++i2)
                        {
                            yield return selector(source[i0, i1, i2], i0, i1, i2);
                        }
                    }
                }
            }
        }
#endif
        public static void ForEach<T>(this IEnumerable<T> source, Action<T> action)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(action, "action");

            foreach (var item in source)
            {
                action(item);
            }
        }
        public static void ForEach<T>(this IEnumerable<T> source, Action<T, int> action)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(action, "action");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    action(item, index);
                    ++index;
                }
            }
        }
        public static string Join(this IEnumerable<string> source, string separator)
        {
            AssertNonNullArgument(source, "source");

            var sb = new StringBuilder();
            source.ForEach((str, index) =>
            {
                if (index == 0)
                    sb.Append(str);
                else
                    sb.Append(separator).Append(str);
            });

            return sb.ToString();
        }
        public static IEnumerable<TResult> SelectWithFirst<T, TResult>(this IEnumerable<T> source, Func<T, T, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(selector, "selector");

            bool isFirst = true;
            var first = default(T);
            foreach (var item in source)
            {
                if (isFirst)
                {
                    first = item;
                    isFirst = false;
                }
                else
                {
                    yield return selector(first, item);
                }
            }
        }

        public static IEnumerable<TResult> ZipWith<T1, T2, TResult>(this IEnumerable<T1> source1, IEnumerable<T2> source2, Func<T1, T2, TResult> func)
        {
            AssertNonNullArgument(source1, "source1");
            AssertNonNullArgument(source2, "source2");
            AssertNonNullArgument(func, "func");

            using (var enumerator1 = source1.GetEnumerator())
            using (var enumerator2 = source2.GetEnumerator())
            {
                while (enumerator1.MoveNext() && enumerator2.MoveNext())
                {
                    yield return func(enumerator1.Current, enumerator2.Current);
                }
            }
        }

        public static IEnumerable<TResult> ZipWith<T1, T2, TResult>(this IEnumerable<T1> source1, IEnumerable<T2> source2, Func<T1, T2, int, TResult> func)
        {
            AssertNonNullArgument(source1, "source1");
            AssertNonNullArgument(source2, "source2");
            AssertNonNullArgument(func, "func");

            using (var enumerator1 = source1.GetEnumerator())
            using (var enumerator2 = source2.GetEnumerator())
            {
                int index = 0;
                while (enumerator1.MoveNext() && enumerator2.MoveNext())
                {
                    yield return func(enumerator1.Current, enumerator2.Current, index);
                    index = checked(index + 1);
                }
            }
        }

        public static IEnumerable<T> Scan<T>(this IEnumerable<T> source, Func<T, T, T> func)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(func, "func");

            var result = default(T);
            bool first = true;
            foreach (var item in source)
            {
                if (first)
                {
                    result = item;
                    first = false;
                }
                else
                {
                    result = func(result, item);
                }
                yield return result;
            }
        }

        public static IEnumerable<TResult> Scan<TSource, TResult>(this IEnumerable<TSource> source, TResult seed, Func<TResult, TSource, TResult> func)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(func, "func");

            var result = seed;
            yield return result;

            foreach (var item in source)
            {
                result = func(result, item);
                yield return result;
            }
        }

        public static IEnumerable<TResult> Scan<TSource, TAccumulate, TResult>(this IEnumerable<TSource> source, TAccumulate seed, Func<TAccumulate, TSource, TAccumulate> func, Func<TAccumulate, TResult> resultSelector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(func, "func");
            AssertNonNullArgument(resultSelector, "resultSelector");

            var result = seed;
            yield return resultSelector(result);

            foreach (var item in source)
            {
                result = func(result, item);
                yield return resultSelector(result);
            }
        }

        public static IEnumerable<T> TakeWhile<T>(this IEnumerable<T> source, params Func<T, int, int, bool>[] predicates)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(predicates, "predicates");
            predicates.ForEach((pred, index)
                => AssertNonNullArgument(pred, string.Format("predicates[{0}]", index)));

            using (var enumerator = source.GetEnumerator())
            {
                var predIndex = 0;
                int itemIndex = -1;
                int itemSubIndex = -1;
                while (enumerator.MoveNext())
                {
                    itemIndex = checked(itemIndex + 1);
                    itemSubIndex = checked(itemSubIndex + 1);
                    while (true)
                    {
                        if (predIndex >= predicates.Length)
                        {
                            yield break;
                        }
                        if (predicates[predIndex](enumerator.Current, itemIndex, itemSubIndex))
                        {
                            // OK. do yield.
                            break;
                        }
                        else
                        {
                            // try next predicate
                            predIndex = checked(predIndex + 1);
                            itemSubIndex = 0;
                        }
                    }
                    yield return enumerator.Current;
                }
            }
        }

        public static IEnumerable<T> TakeWhileAndTheLastOne<T>(this IEnumerable<T> source, Func<T, bool> predicate)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(predicate, "predicate");

            return source.TakeWhile((value, _, __) => predicate(value), (_, __, subIndex) => subIndex < 1);
        }

        /// <see>http://d.hatena.ne.jp/siokoshou/20071227#c1198746342</see>
        public static IEnumerable<T> CascadeDepthFirst<T>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func)
        {
            AssertNonNullArgument(source, "source");

            var enumeratorStack = new Stack<IEnumerator<T>>();
            try
            {
                var top = source.GetEnumerator();
                while (true)
                {
                    if (top.MoveNext())
                    {
                        yield return top.Current;
                        enumeratorStack.Push(top);
                        top = func(top.Current).GetEnumerator();
                    }
                    else
                    {
                        if (enumeratorStack.Count <= 0)
                        {
                            yield break;
                        }
                        top = enumeratorStack.Pop();
                    }
                }
            }
            finally
            {
                enumeratorStack.ForEach(e => e.Dispose());
            }
        }
        public static IEnumerable<T> CascadeDepthFirst<T>(this IEnumerable<T> source, ILookup<T, T> lookup)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeDepthFirst(lookup.AsFunc());
        }
        public static IEnumerable<TResult> CascadeDepthFirst<T, TResult>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            var enumeratorStack = new Stack<IEnumerator<T>>();
            try
            {
                var top = source.GetEnumerator();
                while (true)
                {
                    if (top.MoveNext())
                    {
                        yield return selector(top.Current, enumeratorStack.Count);
                        enumeratorStack.Push(top);
                        top = func(top.Current).GetEnumerator();
                    }
                    else
                    {
                        if (enumeratorStack.Count <= 0)
                        {
                            yield break;
                        }
                        top = enumeratorStack.Pop();
                    }
                }

            }
            finally
            {
                enumeratorStack.ForEach(e => e.Dispose());
            }
        }
        public static IEnumerable<TResult> CascadeDepthFirst<T, TResult>(this IEnumerable<T> source, ILookup<T, T> lookup, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeDepthFirst(lookup.AsFunc(), selector);
        }
        public static IEnumerable<T> CascadeBreadthFirst<T>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func)
        {
            AssertNonNullArgument(source, "source");

            var list = source.ToList();
            int nestLevel = 0;
            while (true)
            {
                if (!list.Any()) yield break;
                foreach (var item in list) yield return item;
                ++nestLevel;
                list = list.SelectMany(item => func(item)).ToList();
            }
        }
        public static IEnumerable<T> CascadeBreadthFirst<T>(this IEnumerable<T> source, ILookup<T, T> lookup)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeBreadthFirst(lookup.AsFunc());
        }

        public static IEnumerable<TResult> CascadeBreadthFirst<T, TResult>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");

            var list = source.ToList();
            int nestLevel = 0;
            while (true)
            {
                if (!list.Any()) yield break;
                foreach (var item in list) yield return selector(item, nestLevel);
                ++nestLevel;
                list = list.SelectMany(item => func(item)).ToList();
            }
        }
        public static IEnumerable<TResult> CascadeBreadthFirst<T, TResult>(this IEnumerable<T> source, ILookup<T, T> lookup, Func<T, int, TResult> selector)
        {
            AssertNonNullArgument(source, "source");
            AssertNonNullArgument(lookup, "lookup");

            return source.CascadeBreadthFirst(lookup.AsFunc(), selector);
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(item);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, string category)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(item, category);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, string> selector)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(selector(item));
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, string> selector, string category)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(selector(item), category);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, int, string> selector)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    System.Diagnostics.Trace.WriteLine(selector(item, index));
                    yield return item;
                    ++index;
                }
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, int, string> selector, string category)
        {
            AssertNonNullArgument(source, "source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    System.Diagnostics.Trace.WriteLine(selector(item, index), category);
                    yield return item;
                    ++index;
                }
            }
        }

        public static void ConsoleWriteLine<T>(this IEnumerable<T> source)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(item);
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, Func<T, string> messageSelector)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(messageSelector(item));
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, string format, Func<T, object[]> argumentsSelector)
        {
            AssertNonNullArgument(source, "source");

            foreach (var item in source)
            {
                Console.WriteLine(format, argumentsSelector(item));
            }
        }

        public static IEnumerable<T> Hook<T>(this IEnumerable<T> source, EnumeratorHook<T> hook)
        {
            AssertNonNullArgument(source, "source");

            return hook.Hook(source);
        }
    }

    [StructLayout(LayoutKind.Auto)]
    public struct EnumeratorHook<T>
    {
        public Action OnBeforeBegin { set; internal get; }
        public Action<IEnumerator<T>> OnAfterBegin { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeReset { set; internal get; }
        public Action<IEnumerator<T>> OnAfterReset { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeGetCurrent { set; internal get; }
        public Action<IEnumerator<T>, T> OnAfterGetCurrent { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeMoveNext { set; internal get; }
        public Action<IEnumerator<T>, bool> OnAfterMoveNext { set; internal get; }
        public Action<IEnumerator<T>> OnBeforeDispose { set; internal get; }
        public Action<IEnumerator<T>> OnAfterDispose { set; internal get; }

        public IEnumerable<T> Hook(IEnumerable<T> source)
        {
            MakeWorld.AssertNonNullArgument(source, "source");

            return new MakeWorld.HookableEnumerable<T>(source, this);
        }
    }

    public static partial class MakeWorld
    {
        internal class HookableEnumerable<T> : IEnumerable<T>
        {
            private readonly IEnumerable<T> _source;
            private readonly EnumeratorHook<T> _hook;

            public HookableEnumerable(IEnumerable<T> source, EnumeratorHook<T> hook)
            {
                this._source = source;
                this._hook = hook;
            }
            public IEnumerator<T> GetEnumerator()
            {
                return new HookableEnumerator<T>(_source, _hook);
            }
            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }
        private class HookableEnumerator<T> : IEnumerator<T>
        {
            private readonly EnumeratorHook<T> _hook;
            private readonly IEnumerator<T> _source;

            public HookableEnumerator(IEnumerable<T> source, EnumeratorHook<T> hook)
            {
                _hook = hook;
                InvokeIfNonNull(_hook.OnBeforeBegin);
                _source = source.GetEnumerator();
                InvokeIfNonNull(_hook.OnAfterBegin, _source);
            }

            public T Current
            {
                get
                {
                    InvokeIfNonNull(_hook.OnBeforeGetCurrent, _source);
                    var current = _source.Current;
                    InvokeIfNonNull(_hook.OnAfterGetCurrent, _source, current);
                    return current;
                }
            }
            public void Dispose()
            {
                InvokeIfNonNull(_hook.OnBeforeDispose, _source);
                _source.Dispose();
                InvokeIfNonNull(_hook.OnAfterDispose, _source);
            }
            object System.Collections.IEnumerator.Current
            {
                get { return this.Current; }
            }
            public bool MoveNext()
            {
                InvokeIfNonNull(_hook.OnBeforeMoveNext, _source);
                bool result = _source.MoveNext();
                InvokeIfNonNull(_hook.OnAfterMoveNext, _source, result);
                return result;
            }
            public void Reset()
            {
                InvokeIfNonNull(_hook.OnBeforeReset, _source);
                _source.Reset();
                InvokeIfNonNull(_hook.OnAfterReset, _source);
            }

            private static void InvokeIfNonNull(Action action)
            {
                if (action == null) return;

                action();
            }
            private static void InvokeIfNonNull<T1>(Action<T1> action, T1 arg)
            {
                if (action == null) return;

                action(arg);
            }
            private static void InvokeIfNonNull<T1, T2>(Action<T1, T2> action, T1 arg1, T2 arg2)
            {
                if (action == null) return;

                action(arg1, arg2);
            }
        }
    }
}